// SpawnParticles(x, y, direction, color)
PORTAL.SpawnParticles = '
    if(global.particles == PARTICLES_OFF) {
        exit;
    }

    var blood;
    repeat(4) {
        blood = instance_create(argument0, argument1, BloodDrop);
        blood.speed = random(6);
        blood.direction = argument2 + random(43) - 22;
        blood.lifetime = 60 + random(51) - 25;

        // Set a white sprite so we can color it properly.
        blood.sprite_index = BloodDropSP;
        blood.image_index = 6;
        blood.color = argument3;
    }
';

// TeleportProjectile(projectile, portal, [loop])
PORTAL.TeleportProjectile = '
    if (variable_local_exists("teleported") == true) {
        if (argument2 == true) {
            if (teleported == true) {
                teleported = false;
                exit;
            }
        } else {
            exit;
        }
    }

    // Look for the (first) complementary portal.
    var complementaryColor, pair;
    complementaryColor = make_color_rgb(255 - color_get_red(argument1.color), 255 - color_get_green(argument1.color), 255 - color_get_blue(argument1.color));
    pair = noone;

    with(PORTAL.Portal) {
        if (color == complementaryColor) {
            pair = id;
            break;
        }
    }

    if (pair != noone) {
        teleported = true;

        // Shoot out of the other portal with your current velocity.
        var tmpSpeed, tmpDir;
        tmpSpeed = speed;
        tmpDir = 180 + argument1.direction - direction + pair.direction;
        hspeed = lengthdir_x(tmpSpeed, tmpDir);
        vspeed = lengthdir_y(tmpSpeed, tmpDir);

        if (object_index != Sentry) {
            image_angle = direction;
        }

        // Dont spawn in the wall.
        x = pair.x + lengthdir_x(15, pair.direction);
        y = pair.y + lengthdir_y(15, pair.direction);
    }
';

// SendPlacePortal(x, y, orientation, dir, color)
PORTAL.SendPlacePortal = '
    var portal;
    write_ubyte(PORTAL.sendBuffer, PORTAL.PLACE_PORTAL);

    write_ushort(PORTAL.sendBuffer, argument0 * 5);
    write_ushort(PORTAL.sendBuffer, argument1 * 5);
    write_ubyte(PORTAL.sendBuffer, argument2);
    write_ubyte(PORTAL.sendBuffer, argument3 / 2);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument4));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument4));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument4));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    portal = execute_string(PORTAL.PlacePortal, argument0, argument1, argument2, argument3, argument4);
    return portal;
';

// instance PlacePortal(x, y, orientation, dir, color)
PORTAL.PlacePortal = '
    var portal;
    portal = instance_create(argument0, argument1, PORTAL.Portal);
    portal.image_index = argument2;
    portal.direction = argument3;
    portal.color = argument4;

    // Link these portals so they are marked on the cursor.
    if (global.myself.object != -1) {
        if (global.myself.object.currentWeapon.object_index == PORTAL.Portalgun) {
            var col;
            col = make_color_rgb(global.myself.object.currentWeapon.colorR, global.myself.object.currentWeapon.colorG, global.myself.object.currentWeapon.colorB);
            if (col == argument4) {
                global.myself.object.currentWeapon.primary = portal;
            } else {
                col = make_color_rgb(255 - global.myself.object.currentWeapon.colorR, 255 - global.myself.object.currentWeapon.colorG, 255 - global.myself.object.currentWeapon.colorB);
                if (col == argument4) {
                    global.myself.object.currentWeapon.secondary = portal;
                }
            }
        }
    }

    return portal;
';

// SendDestroyPortal(portal)
PORTAL.SendDestroyPortal = '
    write_ubyte(PORTAL.sendBuffer, PORTAL.DESTROY_PORTAL);

    write_ubyte(PORTAL.sendBuffer, ds_list_find_index(PORTAL.portals, argument0));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.DestroyPortal, ds_list_find_index(PORTAL.portals, argument0));
';

// DestroyPortal(portalId)
PORTAL.DestroyPortal = '
    var portal;
    portal = ds_list_find_value(PORTAL.portals, argument0);

    with(portal) {
        instance_destroy();
    }
';

// SendPlacePedestal(x, y, color, respawnTime)
PORTAL.SendPlacePedestal = '
    var pedestal;
    write_ubyte(PORTAL.sendBuffer, PORTAL.PLACE_PEDESTAL);

    write_ushort(PORTAL.sendBuffer, argument0 * 5);
    write_ushort(PORTAL.sendBuffer, argument1 * 5);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument2));
    write_ushort(PORTAL.sendBuffer, argument3);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    pedestal = execute_string(PORTAL.PlacePedestal, argument0, argument1, argument2, argument3);
    return pedestal;
';

// instance PlacePedestal(x, y, color, respawnTime)
PORTAL.PlacePedestal = '
    var pedestal;
    pedestal = instance_create(argument0, argument1, PORTAL.Pedestal);
    pedestal.defaultColor = argument2;
    pedestal.color = argument2;
    pedestal.respawnTime = argument3;
    pedestal.ownedByMap = false;

    return pedestal;
';

// SendEquip(player, pedestal, color)
PORTAL.SendEquip = '
    var playerId, pedestalId;
    playerId =  ds_list_find_index(global.players, argument0);
    if (argument1 > 0) {
        pedestalId =  ds_list_find_index(PORTAL.pedestals, argument1);
    } else {
        pedestalId = 255;
    }

    write_ubyte(PORTAL.sendBuffer, PORTAL.EQUIP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, pedestalId);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument2));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Equip, playerId, pedestalId, argument2);
';

// Equip(playerId, pedestalId, color)
PORTAL.Equip = '
    var player, pedestal;
    player = ds_list_find_value(global.players, argument0);

    if (player.object == -1) {
        show_error("Giving a portalgun to a dead player GG.", false);
        exit;
    }

    if (argument1 != 255) {
        pedestal = ds_list_find_value(PORTAL.pedestals, argument1);

        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            pedestal.color = make_color_rgb(player.object.currentWeapon.colorR, player.object.currentWeapon.colorG, player.object.currentWeapon.colorB);
        } else {
            pedestal.hasPortalgun = false;
            player.object.sourcePedestal = pedestal;
            if (pedestal.respawnTime > 0) {
                with(pedestal) {
                    alarm[0] = respawnTime / global.delta_factor;
                }
            }
        }
    }

    with(player.object.currentWeapon) {
        instance_destroy();
    }

    global.paramOwner = player.object;
    player.object.currentWeapon = instance_create(x, y, PORTAL.Portalgun);
    global.paramOwner = noone;

    player.object.currentWeapon.colorR = color_get_red(argument2);
    player.object.currentWeapon.colorG = color_get_green(argument2);
    player.object.currentWeapon.colorB = color_get_blue(argument2);
';

// SendUnequip(player, pedestal)
PORTAL.SendUnequip = '
    var playerId, pedestalId;
    playerId = ds_list_find_index(global.players, argument0);
    if (argument1 > 0) {
        pedestalId =  ds_list_find_index(PORTAL.pedestals, argument1);
    } else {
        pedestalId = 255;
    }

    write_ubyte(PORTAL.sendBuffer, PORTAL.UNEQUIP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, pedestalId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Unequip, playerId, pedestalId);
';

// Unequip(playerId, pedestalId)
PORTAL.Unequip = '
    var player, pedestal;
    player = ds_list_find_value(global.players, argument0);

    if (player.object == -1) {
        show_error("Taking a portalgun from a dead player GG.", false);
        exit;
    }

    if (argument1 != 255) {
        pedestal = ds_list_find_value(PORTAL.pedestals, argument1);

        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            player.object.sourcePedestal = noone;
            with(pedestal) {
                color = make_color_rgb(player.object.currentWeapon.colorR, player.object.currentWeapon.colorG, player.object.currentWeapon.colorB);
                alarm[0] = -1;
                hasPortalgun = true;
            }
        } else {
            show_error("Unequipping non-existing portalguns...", false);
            exit;
        }
    }

    with(player.object.currentWeapon) {
        instance_destroy();
    }

    global.paramOwner = player.object;
    player.object.currentWeapon = instance_create(x, y, player.object.weapons[0]);
    global.paramOwner = noone;
';

// SendFullState()
PORTAL.SendFullState = '
    var i, pedestal, portal;

    write_ubyte(PORTAL.sendBuffer, PORTAL.FULL_STATE);
    write_ubyte(PORTAL.sendBuffer, ds_list_size(PORTAL.pedestals));
    for (i = 0; i < ds_list_size(PORTAL.pedestals); i+= 1) {
        pedestal = ds_list_find_value(PORTAL.pedestals, i);
        if (pedestal.ownedByMap == false) {
            write_ushort(PORTAL.sendBuffer, pedestal.x * 5);
            write_ushort(PORTAL.sendBuffer, pedestal.y * 5);
        }

        write_ubyte(PORTAL.sendBuffer, pedestal.hasPortalgun);
        write_ushort(PORTAL.sendBuffer, pedestal.respawnTime);
        write_ushort(PORTAL.sendBuffer, pedestal.alarm[0] * global.delta_factor);
    }

    write_ubyte(PORTAL.sendBuffer, ds_list_size(PORTAL.portals));
    for (i = 0; i < ds_list_size(PORTAL.portals); i+= 1) {
        portal = ds_list_find_value(PORTAL.portals, i);
        write_ushort(PORTAL.sendBuffer, portal.x * 5);
        write_ushort(PORTAL.sendBuffer, portal.y * 5);
        write_ubyte(PORTAL.sendBuffer, portal.image_index);
        write_ubyte(PORTAL.sendBuffer, portal.direction / 2);
        write_ubyte(PORTAL.sendBuffer, color_get_red(portal.color));
        write_ubyte(PORTAL.sendBuffer, color_get_green(portal.color));
        write_ubyte(PORTAL.sendBuffer, color_get_blue(portal.color));
    }

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);
';

// ReadFullStata(buffer)
PORTAL.ReadFullState = '
    var i, count, pedestal, portal;

    count = read_ubyte(argument0);
    for (i = 0; i < count; i+= 1) {
        if (ds_list_size(PORTAL.pedestals) > i) {
            pedestal = ds_list_find_value(PORTAL.pedestals, i);
        } else {
            pedestal = instance_create(read_ushort(argument0) / 5, read_ushort(argument0) / 5, PORTAL.Pedestal);
        }

        pedestal.hasPortalgun = read_ubyte(argument0);
        pedestal.respawnTime = read_ushort(argument0);
        pedestal.alarm[0] = read_ushort(argument0) / global.delta_factor;
    }

    count = read_ubyte(argument0);
    for (i = 0; i < count; i+= 1) {
        portal = instance_create(read_ushort(argument0) / 5, read_ushort(argument0) / 5, PORTAL.Portal);
        portal.image_index = read_ubyte(argument0);
        portal.direction = read_ubyte(argument0) * 2;
        portal.color = make_color_rgb(read_ubyte(argument0), read_ubyte(argument0), read_ubyte(argument0));
    }
';